----------------------------------------------------------------------------------
-- Company:             www.kampis-elektroecke.de
-- Engineer:            Daniel Kampert
-- 
-- Create Date:         30.01.2018 15:54:57
-- Design Name: 
-- Module Name:         I2S_Top - I2S_Top_Arch
-- Project Name: 
-- Target Devices:      XC7Z010CLG400-1
-- Tool Versions:       Vivado 2016.4
-- Description: 
-- 
-- Dependencies: 
-- 
-- Revision:
--  Revision 0.01   30.01.2018      File Created
--
-- Additional Comments:
--  > ROM needs one cycle to read data
----------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx leaf cells in this code.
library UNISIM;
use UNISIM.VComponents.all;

entity I2S_Top is
    Generic ( DATA_WIDTH_PER_CHANNEL : INTEGER := 16;
              MCLK_PRESCALER : INTEGER := 256 
              );
    Port (  Clock    : in STD_LOGIC;
            Reset    : in STD_LOGIC;
            MCLK     : out STD_LOGIC;
            LRCLK    : out STD_LOGIC;
            I2S_SCLK : out STD_LOGIC;
            I2S_Data : out STD_LOGIC
            );
end I2S_Top;

architecture I2S_Top_Arch of I2S_Top is

    component I2S_Transmitter is
        generic (   DATA_WIDTH_PER_CHANNEL : INTEGER := 16;
                    MCLK_PRESCALER : INTEGER := 256
                    );
        port (  Clock       :   in STD_LOGIC;
                Reset       :   in STD_LOGIC;
                Data        :   in STD_LOGIC_VECTOR(((DATA_WIDTH_PER_CHANNEL * 2) - 1) downto 0);
                Strobe      :   in STD_LOGIC;
                Busy        :   out STD_LOGIC;
                MCLK        :   out STD_LOGIC;
                LRCLK       :   out STD_LOGIC;
                I2S_SCLK    :   out STD_LOGIC;
                I2S_Data    :   out STD_LOGIC
                );
    end component I2S_Transmitter;

    component Sine_ROM is
        Port (  Clock   :   in STD_LOGIC;
                Reset   :   in STD_LOGIC;
                Address :   in STD_LOGIC_VECTOR(6 downto 0);
                Data_Out:   out STD_LOGIC_VECTOR(15 downto 0)
                );
    end component;
    
    component ClockManager is
        Port (  Clock_In : in STD_LOGIC;
                I2S : out STD_LOGIC;
                Locked : out STD_LOGIC;
                Reset : in STD_LOGIC
                );
    end component ClockManager;

    -- Variables for state machine
    -- State definition
    type STATES is(ResetState, Z0, Z1);
    
    signal CurrentState :   STATES := ResetState;

    -- Signals for I2S-Transmitter
    signal Strobe       :   STD_LOGIC := '0';
    signal Busy         :   STD_LOGIC := '0';
    signal Locked       :   STD_LOGIC := '0';
    signal TransmitData :   STD_LOGIC_VECTOR(0 to ((DATA_WIDTH_PER_CHANNEL * 2) - 1)) := (others => '0');
    
    -- Signal for Audio ROM
    signal ROM_Data     :   STD_LOGIC_VECTOR(0 to (DATA_WIDTH_PER_CHANNEL - 1)) := (others => '0');
    signal ROM_Address  :   STD_LOGIC_VECTOR(6 downto 0) := (others => '0');

begin

    Transmitter : I2S_Transmitter generic map ( DATA_WIDTH_PER_CHANNEL => DATA_WIDTH_PER_CHANNEL,
                                                MCLK_PRESCALER => MCLK_PRESCALER)
                                  port map ( Clock => Clock,
                                             Reset => Reset,
                                             Data => TransmitData,
                                             Strobe => Strobe,
                                             Busy => Busy,
                                             MCLK => MCLK,
                                             LRCLK => LRCLK,
                                             I2S_SCLK => I2S_SCLK,
                                             I2S_Data => I2S_Data
                                             );

     Audio_Data : Sine_ROM port map (   Clock => Clock,
                                        Reset => Reset,
                                        Address => ROM_Address,
                                        Data_Out => ROM_Data
                                        );

    -- State network
    StateChange : process(Clock, Reset, CurrentState)
        variable NextState : STATES;
    begin
        if(Reset = '1') then
            NextState := ResetState;
        elsif(rising_Edge(Clock)) then
            case CurrentState is
                when ResetState =>  if(Reset = '1') then
                                        NextState := ResetState;
                                    else
                                        NextState := Z0;
                                    end if;
                when Z0 =>  if(Busy = '1') then
                                NextState := Z1;
                            else
                                NextState := Z0;
                            end if;
                when Z1 =>  if(Busy = '0') then
                                NextState := Z0;
                            end if;
            end case;
        end if;
        
        CurrentState <= NextState;
        
    end process;
    
    -- Output network
    Output : process(CurrentState)
        variable Address : INTEGER;
    begin
        case CurrentState is
            when ResetState => Address := 0;
            when Z0 =>  Strobe <= '1';
            when Z1 =>  Strobe <= '0';
                        if(Address = 99) then
                            Address := 0;
                        else
                            Address := Address + 1;
                        end if;
        end case;
        
        ROM_Address <= STD_LOGIC_VECTOR(to_unsigned(Address, ROM_Address'length));        
    end process;
    
    TransmitData <= ROM_Data & ROM_Data;
    
end I2S_Top_Arch;