----------------------------------------------------------------------------------
-- Company:             www.kampis-elektroecke.de
-- Engineer:            Daniel Kampert
-- 
-- Create Date:         11.05.2016 09:04:43
-- Design Name: 
-- Module Name:         I2S_Transmitter - I2S_Transmitter_Arch
-- Project Name: 
-- Target Devices:      XC7Z010CLG400-1
-- Tool Versions:       Vivado 2016.4
-- Description:         Simple I2S-Transmitter with master clock for I2S device
-- 
-- Dependencies: 
-- 
-- Revision:
--  Revision 0.01   11.05.2016      File Created
--  Revision 1.00   30.05.2016      First release
--  Revision 2.00   08.02.2018      Rework the code and add some handshakeing
--
-- Known issues:
--  > First high bit after reset will not be send
--
-- Additional Comments:
--      > Reset must hold high for 4 i2s clock cycles after Busy = 0 to make sure that all bits 
--        of the last message are transmitted
--      > Strobe must be high at least for 2 i2s clock cycles
----------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx leaf cells in this code.
library UNISIM;
use UNISIM.VComponents.all;

entity I2S_Transmitter is
    generic (   DATA_WIDTH_PER_CHANNEL : INTEGER := 16;
                MCLK_PRESCALER : INTEGER := 256
                );
    port (  Clock       :   in STD_LOGIC;
            Reset       :   in STD_LOGIC;
            Data        :   in STD_LOGIC_VECTOR(((DATA_WIDTH_PER_CHANNEL * 2) - 1) downto 0);
            Strobe      :   in STD_LOGIC;
            Busy        :   out STD_LOGIC;
            MCLK        :   out STD_LOGIC;
            LRCLK       :   out STD_LOGIC;
            I2S_SCLK    :   out STD_LOGIC;
            I2S_Data    :   out STD_LOGIC
            );
end entity;

architecture I2S_Transmitter_Arch of I2S_Transmitter is

    -- Clock divider for MCLK and I2S clock
    component ClockDivider is
        Generic ( PRESCALER : INTEGER
                );
        Port ( Clock_In : in STD_LOGIC;
               Reset : in STD_LOGIC;
               Clock_Out : out STD_LOGIC
               );
    end component;

    -- Buffer for channel data
    signal DataBuffer   :   STD_LOGIC_VECTOR(0 to ((DATA_WIDTH_PER_CHANNEL * 2) - 1)) := (others => '0');
    
    signal I2S_Data_Int :   STD_LOGIC_VECTOR(1 downto 0) := (others => '0');
    signal LRCLK_Int    :   STD_LOGIC := '0';
    signal MCLK_Int     :   STD_LOGIC := '0';
    signal Busy_Int     :   STD_LOGIC := '0';
    signal Clock_Int    :   STD_LOGIC := '0';
    signal EnableClock  :   STD_LOGIC := '0';
    signal Counter      :   INTEGER range 0 to ((DATA_WIDTH_PER_CHANNEL * 2) - 1);
    signal CounterDelay :   INTEGER range 0 to ((DATA_WIDTH_PER_CHANNEL * 2) - 1);
    signal CounterDelay2:   INTEGER range 0 to ((DATA_WIDTH_PER_CHANNEL * 2) - 1);

begin

    -- Generate the I2S clock
    I2S_Clock_Divider : ClockDivider generic map (Prescaler => (MCLK_PRESCALER / (2 * DATA_WIDTH_PER_CHANNEL)))
                                port map (  Clock_In => Clock,
                                            Reset => Reset,
                                            Clock_Out => Clock_Int
                                            );

    -- Generate the MCLK (just a single CLB to isolate the clock from output)
    MCLK_Inst : ClockDivider  generic map (Prescaler => 1)
                                port map (  Clock_In => Clock,
                                            Reset => Reset,
                                            Clock_Out => MCLK_Int
                                            );

Load_data : process(Clock_Int, Reset, Strobe, Data)
    begin
        if(Reset = '1') then
            DataBuffer <= (others => '0');
        elsif(falling_edge(Clock_Int)) then
            if(Strobe = '1') then
                DataBuffer <= Data;
            end if;
        end if;
    end process;
 
Busy_Logic : process(Clock_Int, Reset, Strobe)
    begin
        if(Reset = '1') then
            Busy_Int <= '0';
        elsif(falling_edge(Clock_Int)) then
            if((Counter > 0) and (Counter < ((DATA_WIDTH_PER_CHANNEL * 2) - 2))) then
                Busy_Int <= '1';
            else
                Busy_Int <= '0';
            end if;
        end if;
    end process;
 
Data_logic : process(Clock_Int, Reset, Counter)
    begin
        if(Reset = '1') then
            I2S_Data_Int(0) <= '0';
        elsif(falling_edge(Clock_Int)) then
            -- Shift out data
            I2S_Data_Int(0) <= DataBuffer(Counter);
        end if;
    end process;

LRCLK_Logic : process(Clock_Int, Reset)
    begin
        if(Reset = '1') then
            LRCLK_Int <= '0';
            EnableClock <= '0';
        elsif(falling_edge(Clock_Int)) then
            if(Counter <= (DATA_WIDTH_PER_CHANNEL - 1)) then
                LRCLK_Int <= '0';
            else
                LRCLK_Int <= '1';
            end if;
            
            -- Delay the clock by one cycle after a reset
            -- to remove the 17th clock cycle during the first LRCLK
            EnableClock <= '1';
            
        end if;
    end process;

BitCounter : process(Clock_Int, Reset)
    begin
        if(Reset = '1') then
            Counter <= 0;
            I2S_Data_Int(1) <= '0';
        elsif(falling_edge(Clock_Int)) then
            if(Counter = ((DATA_WIDTH_PER_CHANNEL * 2) - 1)) then
                Counter <= 0;
            else
                Counter <= Counter + 1;  
            end if; 

            -- Delay the signals by one      
            CounterDelay <= Counter;
            CounterDelay2 <= CounterDelay;
            I2S_Data_Int(1) <= I2S_Data_Int(0);   
        end if;
    end process; 

    Busy <= Busy_Int;
    LRCLK <= LRCLK_Int;
    MCLK <= MCLK_Int and EnableClock;
    I2S_SCLK <= Clock_Int and EnableClock;
    I2S_Data <= I2S_Data_Int(1);

end architecture I2S_Transmitter_Arch;